# ------------------------------------------------------------------------
#
# ------------------------------------------------------------------------

'''

@ARTICLE{10250890,
  author={Guo, Yujuan and Fu, Xiyou and Xu, Meng and Jia, Sen},
  journal={IEEE Transactions on Geoscience and Remote Sensing}, 
  title={Stereo Cross-Attention Network for Unregistered Hyperspectral and Multispectral Image Fusion}, 
  year={2023},
  volume={61},
  number={},
  pages={1-15},
  doi={10.1109/TGRS.2023.3314755}}
'''

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


##########################################################################################
class Local_Base():
    def convert(self, *args, train_size, **kwargs):
        replace_layers(self, *args, train_size=train_size, **kwargs)
        imgs = torch.rand(train_size)
        with torch.no_grad():
            self.forward(imgs)
def replace_layers(model, base_size, train_size, fast_imp, **kwargs):
    for n, m in model.named_children():
        if len(list(m.children())) > 0:
            ## compound module, go inside it
            replace_layers(m, base_size, train_size, fast_imp, **kwargs)

        if isinstance(m, nn.AdaptiveAvgPool2d):
            pool = AvgPool2d(base_size=base_size, fast_imp=fast_imp, train_size=train_size)
            assert m.output_size == 1
            setattr(model, n, pool)

class AvgPool2d(nn.Module):
    def __init__(self, kernel_size=None, base_size=None, auto_pad=True, fast_imp=False, train_size=None):
        super().__init__()
        self.kernel_size = kernel_size
        self.base_size = base_size
        self.auto_pad = auto_pad

        # only used for fast implementation
        self.fast_imp = fast_imp
        self.rs = [5, 4, 3, 2, 1]
        self.max_r1 = self.rs[0]
        self.max_r2 = self.rs[0]
        self.train_size = train_size

    def extra_repr(self) -> str:
        return 'kernel_size={}, base_size={}, stride={}, fast_imp={}'.format(
            self.kernel_size, self.base_size, self.kernel_size, self.fast_imp
        )

    def forward(self, x):
        if self.kernel_size is None and self.base_size:
            train_size = self.train_size
            if isinstance(self.base_size, int):
                self.base_size = (self.base_size, self.base_size)
            self.kernel_size = list(self.base_size)
            self.kernel_size[0] = x.shape[2] * self.base_size[0] // train_size[-2]
            self.kernel_size[1] = x.shape[3] * self.base_size[1] // train_size[-1]

            # only used for fast implementation
            self.max_r1 = max(1, self.rs[0] * x.shape[2] // train_size[-2])
            self.max_r2 = max(1, self.rs[0] * x.shape[3] // train_size[-1])

        if self.kernel_size[0] >= x.size(-2) and self.kernel_size[1] >= x.size(-1):
            return F.adaptive_avg_pool2d(x, 1)

        if self.fast_imp:  # Non-equivalent implementation but faster
            h, w = x.shape[2:]
            if self.kernel_size[0] >= h and self.kernel_size[1] >= w:
                out = F.adaptive_avg_pool2d(x, 1)
            else:
                r1 = [r for r in self.rs if h % r == 0][0]
                r2 = [r for r in self.rs if w % r == 0][0]
                # reduction_constraint
                r1 = min(self.max_r1, r1)
                r2 = min(self.max_r2, r2)
                s = x[:, :, ::r1, ::r2].cumsum(dim=-1).cumsum(dim=-2)
                n, c, h, w = s.shape
                k1, k2 = min(h - 1, self.kernel_size[0] // r1), min(w - 1, self.kernel_size[1] // r2)
                out = (s[:, :, :-k1, :-k2] - s[:, :, :-k1, k2:] - s[:, :, k1:, :-k2] + s[:, :, k1:, k2:]) / (k1 * k2)
                out = torch.nn.functional.interpolate(out, scale_factor=(r1, r2))
        else:
            n, c, h, w = x.shape
            s = x.cumsum(dim=-1).cumsum_(dim=-2)
            s = torch.nn.functional.pad(s, (1, 0, 1, 0))  # pad 0 for convenience
            k1, k2 = min(h, self.kernel_size[0]), min(w, self.kernel_size[1])
            s1, s2, s3, s4 = s[:, :, :-k1, :-k2], s[:, :, :-k1, k2:], s[:, :, k1:, :-k2], s[:, :, k1:, k2:]
            out = s4 + s1 - s2 - s3
            out = out / (k1 * k2)

        if self.auto_pad:
            n, c, h, w = x.shape
            _h, _w = out.shape[2:]
            # print(x.shape, self.kernel_size)
            pad2d = ((w - _w) // 2, (w - _w + 1) // 2, (h - _h) // 2, (h - _h + 1) // 2)
            out = torch.nn.functional.pad(out, pad2d, mode='replicate')

        return out


class LayerNorm2d(nn.Module):

    def __init__(self, channels, eps=1e-6):
        super(LayerNorm2d, self).__init__()
        self.register_parameter('weight', nn.Parameter(torch.ones(channels)))
        self.register_parameter('bias', nn.Parameter(torch.zeros(channels)))
        self.eps = eps

    def forward(self, x):
        return LayerNormFunction.apply(x, self.weight, self.bias, self.eps)

class LayerNormFunction(torch.autograd.Function):

    @staticmethod
    def forward(ctx, x, weight, bias, eps):
        ctx.eps = eps
        N, C, H, W = x.size()
        mu = x.mean(1, keepdim=True)
        var = (x - mu).pow(2).mean(1, keepdim=True)
        y = (x - mu) / (var + eps).sqrt()
        ctx.save_for_backward(y, var, weight)
        y = weight.view(1, C, 1, 1) * y + bias.view(1, C, 1, 1)
        return y

    @staticmethod
    def backward(ctx, grad_output):
        eps = ctx.eps

        N, C, H, W = grad_output.size()
        y, var, weight = ctx.saved_variables
        g = grad_output * weight.view(1, C, 1, 1)
        mean_g = g.mean(dim=1, keepdim=True)

        mean_gy = (g * y).mean(dim=1, keepdim=True)
        gx = 1. / torch.sqrt(var + eps) * (g - y * mean_gy - mean_g)
        return gx, (grad_output * y).sum(dim=3).sum(dim=2).sum(dim=0), grad_output.sum(dim=3).sum(dim=2).sum(
            dim=0), None

# handle multiple input
class MySequential(nn.Sequential):
    def forward(self, *inputs):
        for module in self._modules.values():
            if type(inputs) == tuple:
                inputs = module(*inputs)
            else:
                inputs = module(inputs)
        return inputs

class SimpleGate(nn.Module):
    def forward(self, x):
        x = F.relu(x)
        x1, x2 = x.chunk(2, dim=1)
        return x1 * x2


class SCFBlock(nn.Module):
    def __init__(self, c, DW_Expand=2, FFN_Expand=2, drop_out_rate=0.):
        super().__init__()
        dw_channel = c * DW_Expand
        self.conv1 = nn.Conv2d(in_channels=c, out_channels=dw_channel, kernel_size=1, padding=0, stride=1, groups=1,
                               bias=True)
        self.conv2 = nn.Conv2d(in_channels=dw_channel, out_channels=dw_channel, kernel_size=3, padding=1, stride=1,
                               groups=dw_channel,
                               bias=True)
        self.conv3 = nn.Conv2d(in_channels=dw_channel // 2, out_channels=c, kernel_size=1, padding=0, stride=1,
                               groups=1, bias=True)

        # Simplified Channel Attention
        self.sca = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels=dw_channel // 2, out_channels=dw_channel // 2, kernel_size=1, padding=0, stride=1,
                      groups=1, bias=True),
        )

        # SimpleGate
        self.sg = SimpleGate()

        ffn_channel = FFN_Expand * c
        self.conv4 = nn.Conv2d(in_channels=c, out_channels=ffn_channel, kernel_size=1, padding=0, stride=1, groups=1,
                               bias=True)
        self.conv5 = nn.Conv2d(in_channels=ffn_channel // 2, out_channels=c, kernel_size=1, padding=0, stride=1,
                               groups=1, bias=True)

        self.norm1 = LayerNorm2d(c)
        self.norm2 = LayerNorm2d(c)

        self.dropout1 = nn.Dropout(drop_out_rate) if drop_out_rate > 0. else nn.Identity()
        self.dropout2 = nn.Dropout(drop_out_rate) if drop_out_rate > 0. else nn.Identity()

        self.beta = nn.Parameter(torch.zeros((1, c, 1, 1)), requires_grad=True)
        self.gamma = nn.Parameter(torch.zeros((1, c, 1, 1)), requires_grad=True)

    def forward(self, inp):
        x = inp

        x = self.norm1(x)

        x = self.conv1(x)
        x = self.conv2(x)
        x = self.sg(x)
        x = x * self.sca(x)
        x = self.conv3(x)

        x = self.dropout1(x)

        y = inp + x * self.beta

        x = self.conv4(self.norm2(y))
        x = self.sg(x)
        x = self.conv5(x)

        x = self.dropout2(x)

        return y + x * self.gamma

##########################################################################################
def INF(B, H, W):
        return -torch.diag(torch.tensor(float("inf")).repeat(H), 0).unsqueeze(0).repeat(B * W, 1, 1)

class CCFA(nn.Module):       ##########十字交叉自注意力
    """ Criss-Cross Attention Module"""

    def __init__(self, c):
        super(CCFA, self).__init__()
        self.scale = c ** -0.5
        self.query_conv = nn.Conv2d(c, c, kernel_size=1)
        self.key_conv = nn.Conv2d(c, c, kernel_size=1)
        self.value_conv = nn.Conv2d(c, c, kernel_size=1)
        self.softmax = nn.Softmax(dim=3)
        self.INF = INF
        self.beta = nn.Parameter(torch.zeros((1, c, 1, 1)), requires_grad=True)
        self.gamma = nn.Parameter(torch.zeros((1, c, 1, 1)), requires_grad=True)

    def forward(self, x_l, x_r):
        m_batchsize, _, height, width = x_l.size()

        proj_query_l = self.query_conv(x_l)
        proj_query_r = self.query_conv(x_r)
        proj_query_l_H = proj_query_l.permute(0, 3, 1, 2).contiguous().\
            view(m_batchsize * width, -1, height).permute(0, 2,1)
        proj_query_l_W = proj_query_l.permute(0, 2, 1, 3).contiguous().\
            view(m_batchsize * height, -1, width).permute(0, 2, 1)
        proj_query_r_H = proj_query_r.permute(0, 3, 1, 2).contiguous(). \
            view(m_batchsize * width, -1, height).permute(0, 2, 1)
        proj_query_r_W = proj_query_r.permute(0, 2, 1, 3).contiguous(). \
            view(m_batchsize * height, -1, width).permute(0, 2, 1)


        proj_key_l = self.key_conv(x_l)
        proj_key_r = self.key_conv(x_r)
        proj_key_l_H = proj_key_l.permute(0, 3, 1, 2).contiguous().view(m_batchsize * width, -1, height)
        proj_key_l_W = proj_key_l.permute(0, 2, 1, 3).contiguous().view(m_batchsize * height, -1, width)
        proj_key_r_H = proj_key_r.permute(0, 3, 1, 2).contiguous().view(m_batchsize * width, -1, height)
        proj_key_r_W = proj_key_r.permute(0, 2, 1, 3).contiguous().view(m_batchsize * height, -1, width)

        proj_value_l = self.value_conv(x_l)
        proj_value_r = self.value_conv(x_r)
        proj_value_l_H = proj_value_l.permute(0, 3, 1, 2).contiguous().view(m_batchsize * width, -1, height)
        proj_value_l_W = proj_value_l.permute(0, 2, 1, 3).contiguous().view(m_batchsize * height, -1, width)
        proj_value_r_H = proj_value_r.permute(0, 3, 1, 2).contiguous().view(m_batchsize * width, -1, height)
        proj_value_r_W = proj_value_r.permute(0, 2, 1, 3).contiguous().view(m_batchsize * height, -1, width)

        energy_l2r_H = (torch.bmm(proj_query_l_H, proj_key_r_W)+self.INF(m_batchsize, height, width)).\
            view(m_batchsize, width, height,height).permute(0,2,1,3) * self.scale
        energy_l2r_W = torch.bmm(proj_query_l_W, proj_key_r_H).view(m_batchsize, height, width, width)* self.scale
        energy_r2l_H = (torch.bmm(proj_query_r_H, proj_key_l_W) + self.INF(m_batchsize, height, width)). \
                           view(m_batchsize, width, height, height).permute(0, 2, 1, 3) * self.scale
        energy_r2l_W = torch.bmm(proj_query_r_W, proj_key_l_H).view(m_batchsize, height, width, width) * self.scale

        concate1 = self.softmax(torch.cat([energy_l2r_H, energy_r2l_W], 3))
        concate2 = self.softmax(torch.cat([energy_r2l_H, energy_l2r_W], 3))

        att1_H = concate1[:, :, :, 0:height].permute(0, 2, 1, 3).contiguous().view(m_batchsize * width, height, height)
        att1_W = concate1[:, :, :, height:height + width].contiguous().view(m_batchsize * height, width, width)
        att2_H = concate2[:, :, :, 0:height].permute(0, 2, 1, 3).contiguous().view(m_batchsize * width, height, height)
        att2_W = concate2[:, :, :, height:height + width].contiguous().view(m_batchsize * height, width, width)

        out_l_H = torch.bmm(proj_value_l_H, att1_H.permute(0, 2, 1)).view(m_batchsize, width, -1, height).permute(0, 2, 3, 1)
        out_l_W = torch.bmm(proj_value_l_W, att1_W.permute(0, 2, 1)).view(m_batchsize, height, -1, width).permute(0, 2, 1, 3)
        out_r_H = torch.bmm(proj_value_r_H, att2_H.permute(0, 2, 1)).view(m_batchsize, width, -1, height).permute(0, 2,
                                                                                                                  3, 1)
        out_r_W = torch.bmm(proj_value_r_W, att2_W.permute(0, 2, 1)).view(m_batchsize, height, -1, width).permute(0, 2,
                                                                                                                  1, 3)
        out_H = self.beta * (out_l_H + out_l_W)
        out_W = self.gamma * (out_r_H + out_r_W)

        return x_l + out_W, x_r + out_H

#############################################################坐标注意力
class h_sigmoid(nn.Module):
    def __init__(self, inplace=True):
        super(h_sigmoid, self).__init__()
        self.relu = nn.ReLU6(inplace=inplace)

    def forward(self, x):
        return self.relu(x + 3) / 6


class h_swish(nn.Module):
    def __init__(self, inplace=True):
        super(h_swish, self).__init__()
        self.sigmoid = h_sigmoid(inplace=inplace)

    def forward(self, x):
        return x * self.sigmoid(x)


class CoordAtt(nn.Module):
    def __init__(self, c):
        super(CoordAtt, self).__init__()
        self.pool_l_h = nn.AdaptiveAvgPool2d((None, 1))
        self.pool_l_w = nn.AdaptiveAvgPool2d((1, None))
        self.pool_r_h = nn.AdaptiveAvgPool2d((None, 1))
        self.pool_r_w = nn.AdaptiveAvgPool2d((1, None))

        mip = max(8, c // 32)

        self.convl = nn.Conv2d(c, mip, kernel_size=1, stride=1, padding=0)
        self.bnl = nn.BatchNorm2d(mip)
        self.act_l = h_swish()
        self.convr = nn.Conv2d(c, mip, kernel_size=1, stride=1, padding=0)
        self.bnr = nn.BatchNorm2d(mip)
        self.act_r = h_swish()

        self.conv_l_h = nn.Conv2d(mip, c, kernel_size=1, stride=1, padding=0)
        self.conv_l_w = nn.Conv2d(mip, c, kernel_size=1, stride=1, padding=0)
        self.conv_r_h = nn.Conv2d(mip, c, kernel_size=1, stride=1, padding=0)
        self.conv_r_w = nn.Conv2d(mip, c, kernel_size=1, stride=1, padding=0)

    def forward(self, x_l, x_r):
        identity_l = x_l
        identity_r = x_r

        n, c, h, w = x_l.size()
        l_h = self.pool_l_h(x_l)
        l_w = self.pool_l_w(x_l).permute(0, 1, 3, 2)
        r_h = self.pool_r_h(x_r)
        r_w = self.pool_r_w(x_r).permute(0, 1, 3, 2)

        y_lcr = torch.cat([l_h, r_w], dim=2)
        y_lcr = self.convl(y_lcr)
        y_lcr = self.bnl(y_lcr)
        y_lcr = self.act_l(y_lcr)
        y_rcl = torch.cat([r_h, l_w], dim=2)
        y_rcl = self.convr(y_rcl)
        y_rcl = self.bnr(y_rcl)
        y_rcl = self.act_r(y_rcl)

        lcr_h, lcr_w = torch.split(y_lcr, [h, w], dim=2)
        lcr_w = lcr_w.permute(0, 1, 3, 2)
        rcl_h, rcl_w = torch.split(y_rcl, [h, w], dim=2)
        rcl_w = rcl_w.permute(0, 1, 3, 2)

        l_h = self.conv_l_h(lcr_h).sigmoid()
        l_w = self.conv_l_w(lcr_w).sigmoid()
        r_h = self.conv_r_h(rcl_h).sigmoid()
        r_w = self.conv_r_w(rcl_w).sigmoid()

        out_l = identity_l * l_w * l_h
        out_r = identity_r * r_w * r_h
        return out_r + x_l, out_l + x_r

##############################################################  DiVCAM
def mean_channels_h(F):
    assert(F.dim() == 4)
    spatial_sum = F.sum(3, keepdim=True)
    return spatial_sum / F.size(3)

def stdv_channels_h(F):
    assert(F.dim() == 4)
    F_mean = mean_channels_h(F)
    F_variance = (F - F_mean).pow(2).sum(3, keepdim=True) / F.size(3)
    return F_variance


def mean_channels_w(F):
    assert(F.dim() == 4)
    spatial_sum = F.sum(2, keepdim=True)
    return spatial_sum / F.size(2)

def stdv_channels_w(F):
    assert(F.dim() == 4)
    F_mean = mean_channels_w(F)
    F_variance = (F - F_mean).pow(2).sum(2, keepdim=True) / F.size(2)
    return F_variance


class DiVCAM(nn.Module):
    def __init__(self, c):
        super().__init__()
        self.scale = c ** -0.5

        self.norm_l = LayerNorm2d(c)
        self.norm_r = LayerNorm2d(c)
        self.l_proj1 = nn.Conv2d(c, c, kernel_size=1, stride=1, padding=0)
        self.r_proj1 = nn.Conv2d(c, c, kernel_size=1, stride=1, padding=0)

        self.beta = nn.Parameter(torch.zeros((1, c, 1, 1)), requires_grad=True)
        self.gamma = nn.Parameter(torch.zeros((1, c, 1, 1)), requires_grad=True)

        self.l_proj2 = nn.Conv2d(c, c, kernel_size=1, stride=1, padding=0)
        self.r_proj2 = nn.Conv2d(c, c, kernel_size=1, stride=1, padding=0)
          #########
        self.contrast_h = stdv_channels_h
        self.contrast_w = stdv_channels_w

        self.conv_h = nn.Conv2d(c, c, kernel_size=1, stride=1, padding=0)
        self.conv_w = nn.Conv2d(c, c, kernel_size=1, stride=1, padding=0)

    def forward(self, x_l, x_r):
        Q_l = self.l_proj1(self.norm_l(x_l)).permute(0, 2, 3, 1)  # B, H, W, c
        K_r_T = self.r_proj1(self.norm_r(x_r)).permute(0, 2, 1, 3)  # B, H, c, W (transposed)

        Q_r = self.l_proj1(self.norm_l(x_l)).permute(0, 2, 3, 1)  # B, H, W, c
        K_l_T = self.r_proj1(self.norm_r(x_r)).permute(0, 2, 1, 3)  # B, H, W, c

        V_l = self.l_proj2(x_l).permute(0, 2, 3, 1)  # B, H, W, c
        V_r = self.r_proj2(x_r).permute(0, 2, 3, 1)  # B, H, W, c

        # (B, H, W, c) x (B, H, c, W) -> (B, H, W, W)
        attention1 = torch.matmul(Q_l, K_r_T) * self.scale

        # x_r_h = self.contrast_h(x_r)
        # x_l_w = self.contrast_w(x_l)
        #
        # a_h = self.conv_h(x_r_h).sigmoid()
        # b_w = self.conv_w(x_l_w).sigmoid()

        attention2 = torch.matmul(Q_r, K_l_T) * self.scale

        F_l2r = torch.matmul(torch.softmax(attention1, dim=-1), V_l)  # B, H, W, c
        F_r2l = torch.matmul(torch.softmax(attention2.permute(0, 2, 3, 1), dim=-1), V_r)  # B, H, W, c


        # scale
        F_r2l = F_r2l.permute(0, 3, 1, 2) * self.beta
        F_l2r = F_l2r.permute(0, 3, 1, 2) * self.gamma


        return x_l + F_r2l, x_r + F_l2r

##########################################################  SAM
from skimage import morphology
def morphologic_process(mask):
    device = mask.device
    b,_,_,_ = mask.shape
    mask = ~mask
    mask_np = mask.cpu().numpy().astype(bool)
    mask_np = morphology.remove_small_objects(mask_np, 20, 2)
    mask_np = morphology.remove_small_holes(mask_np, 10, 2)
    for idx in range(b):
        buffer = np.pad(mask_np[idx,0,:,:],((3,3),(3,3)),'constant')
        buffer = morphology.binary_closing(buffer, morphology.disk(3))
        mask_np[idx,0,:,:] = buffer[3:-3,3:-3]
    mask_np = 1-mask_np
    mask_np = mask_np.astype(float)
    return torch.from_numpy(mask_np).float().to(device)
class RB(nn.Module):
    def __init__(self, channels):
        super(RB, self).__init__()
        self.body = nn.Sequential(
            nn.Conv2d(channels, channels, 3, 1, 1, bias=False),
            nn.ReLU(inplace=True),
            nn.Conv2d(channels, channels, 3, 1, 1, bias=False),
        )
    def forward(self, x):
        out = self.body(x)
        return out + x
class SAM(nn.Module):# stereo attention block
    def __init__(self, c):
        super(SAM, self).__init__()
        self.b1 = nn.Conv2d(c, c, 1, 1, 0, bias=True)
        self.b2 = nn.Conv2d(c, c, 1, 1, 0, bias=True)
        self.rb = RB(c)
        self.softmax = nn.Softmax(-1)
        self.bottleneck = nn.Conv2d(c * 2+1, c, 1, 1, 0, bias=True)
    def forward(self, x_left, x_right):# B * C * H * W
        b, c, h, w = x_left.shape
        buffer_left = self.rb(x_left)
        buffer_right = self.rb(x_right)
        ### M_{right_to_left}
        Q = self.b1(buffer_left).permute(0, 2, 3, 1)  # B * H * W * C
        S = self.b2(buffer_right).permute(0, 2, 1, 3)  # B * H * C * W
        score = torch.bmm(Q.contiguous().view(-1, w, c),
                          S.contiguous().view(-1, c, w))  # (B*H) * W * W
        M_right_to_left = self.softmax(score)

        score_T = score.permute(0,2,1)
        M_left_to_right = self.softmax(score_T)

        # valid mask
        V_left_to_right = torch.sum(M_left_to_right.detach(), 1) > 0.1
        V_left_to_right = V_left_to_right.view(b, 1, h, w)  # B * 1 * H * W
        V_left_to_right = morphologic_process(V_left_to_right)
        V_right_to_left = torch.sum(M_right_to_left.detach(), 1) > 0.1
        V_right_to_left = V_right_to_left.view(b, 1, h, w)  # B * 1 * H * W
        V_right_to_left = morphologic_process(V_right_to_left)

        buffer_R = x_right.permute(0, 2, 3, 1).contiguous().view(-1, w, c)  # (B*H) * W * C
        buffer_l = torch.bmm(M_right_to_left, buffer_R).contiguous().view(b, h, w, c).permute(0, 3, 1, 2)  # B * C * H * W

        buffer_L = x_left.permute(0, 2, 3, 1).contiguous().view(-1, w, c)  # (B*H) * W * C
        buffer_r = torch.bmm(M_left_to_right, buffer_L).contiguous().view(b, h, w, c).permute(0, 3, 1, 2)  # B * C * H * W

        out_L = self.bottleneck(torch.cat((buffer_l, x_left, V_left_to_right), 1))
        out_R = self.bottleneck(torch.cat((buffer_r, x_right, V_right_to_left), 1))

        return out_L, out_R
###############################################################
class SCAM(nn.Module):
    '''
    Stereo Cross Attention Module (SCAM)
    '''
    def __init__(self, c):
        super().__init__()
        self.scale = c ** -0.5

        self.norm_l = LayerNorm2d(c)
        self.norm_r = LayerNorm2d(c)
        self.l_proj1 = nn.Conv2d(c, c, kernel_size=1, stride=1, padding=0)
        self.r_proj1 = nn.Conv2d(c, c, kernel_size=1, stride=1, padding=0)
        
        self.beta = nn.Parameter(torch.zeros((1, c, 1, 1)), requires_grad=True)
        self.gamma = nn.Parameter(torch.zeros((1, c, 1, 1)), requires_grad=True)

        self.l_proj2 = nn.Conv2d(c, c, kernel_size=1, stride=1, padding=0)
        self.r_proj2 = nn.Conv2d(c, c, kernel_size=1, stride=1, padding=0)

    def forward(self, x_l, x_r):
        Q_l = self.l_proj1(self.norm_l(x_l)).permute(0, 2, 3, 1)  # B, H, W, c
        Q_r_T = self.r_proj1(self.norm_r(x_r)).permute(0, 2, 1, 3) # B, H, c, W (transposed)

        V_l = self.l_proj2(x_l).permute(0, 2, 3, 1)  # B, H, W, c
        V_r = self.r_proj2(x_r).permute(0, 2, 3, 1)  # B, H, W, c

        # (B, H, W, c) x (B, H, c, W) -> (B, H, W, W)
        attention = torch.matmul(Q_l, Q_r_T) * self.scale

        F_r2l = torch.matmul(torch.softmax(attention, dim=-1), V_r)  #B, H, W, c
        F_l2r = torch.matmul(torch.softmax(attention.permute(0, 1, 3, 2), dim=-1), V_l) #B, H, W, c

        # scale
        F_r2l = F_r2l.permute(0, 3, 1, 2) * self.beta
        F_l2r = F_l2r.permute(0, 3, 1, 2) * self.gamma
        return x_l + F_r2l, x_r + F_l2r

class DropPath(nn.Module):
    def __init__(self, drop_rate, module):
        super().__init__()
        self.drop_rate = drop_rate
        self.module = module

    def forward(self, *feats):
        if self.training and np.random.rand() < self.drop_rate:
            return feats

        new_feats = self.module(*feats)
        factor = 1. / (1 - self.drop_rate) if self.training else 1.

        if self.training and factor != 1.:
            new_feats = tuple([x+factor*(new_x-x) for x, new_x in zip(feats, new_feats)])
        return new_feats

class SCFBlockSR(nn.Module):
    '''
    SCFBlock for Super-Resolution
    '''
    def __init__(self, c, fusion=False, drop_out_rate=0.):
        super().__init__()
        self.blk = SCFBlock(c, drop_out_rate=drop_out_rate)
        self.fusion = CCFA(c) if fusion else None

    def forward(self, *feats):
        feats = tuple([self.blk(x) for x in feats])
        if self.fusion:
            feats = self.fusion(*feats)
        return feats

class SCANet(nn.Module):
    '''
    SCANet for FUSION
    '''
    def __init__(self, scale_ratio, n_select_bands, n_bands, width=48, num_blks=8, drop_path_rate=0., drop_out_rate=0.1, fusion_from=-1, fusion_to=1000, dual=True):
        super(SCANet, self).__init__()
        self.scale_ratio = scale_ratio
        self.n_bands = n_bands
        self.n_select_bands = n_select_bands
        self.dual = dual    # dual input for stereo SR (left view, right view)
        self.bandup = nn.Conv2d(in_channels=self.n_select_bands, out_channels=n_bands, kernel_size=3, padding=1, stride=1,
                               groups=1,
                               bias=True)
        self.intro = nn.Conv2d(in_channels=self.n_bands, out_channels=width, kernel_size=3, padding=1, stride=1, groups=1,
                              bias=True)
        self.body = MySequential(
            *[DropPath(
                drop_path_rate, 
                SCFBlockSR(
                    width, 
                    fusion=(fusion_from <= i and i <= fusion_to), 
                    drop_out_rate=drop_out_rate
                )) for i in range(num_blks)]
        )

        self.up = nn.Sequential(
            nn.Conv2d(in_channels=width, out_channels=self.n_bands, kernel_size=3, padding=1, stride=1, groups=1, bias=True),
            # nn.PixelShuffle(up_scale)
        )
        self.up_scale = scale_ratio

        self.conv_spat = nn.Sequential(
            nn.Conv2d(n_bands, n_bands, kernel_size=3, stride=1, padding=1),
            nn.ReLU(),
            nn.Conv2d(n_bands, n_bands, kernel_size=3, stride=1, padding=1),
        )
        self.conv_spec = nn.Sequential(
            nn.Conv2d(n_bands, n_bands, kernel_size=3, stride=1, padding=1),
        )


    def forward(self, x_lr, x_hr):    ## inp
        #inp_hr = F.interpolate(inp, scale_factor=self.up_scale, mode='bilinear')

        x_lr = F.interpolate(x_lr, scale_factor=self.scale_ratio, mode='bilinear')
        x_hr = self.bandup(x_hr)
        inp = torch.cat([x_lr, x_hr], dim=0)
        if self.dual:
            inp = inp.chunk(2, dim=0)
        else:
            inp = (inp, )
        feats = [self.intro(x) for x in inp]
        feats = self.body(*feats)
        fea_lr, fea_hr = [self.up(x) for x in feats]
        # x = torch.cat([fea_lr, fea_hr], dim=1)
        x = 0.5*(fea_lr + x_lr) + 0.5*(fea_hr + x_hr)

        return x, 0, 0, 0, 0, 0






